/* lock.c - functions for locking to a TV signal without sync pulses
 *
 * Copyright (C) 2003 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: lock.c 59 2004-07-30 21:47:15Z rsnel $
 */
#include "lock.h"
#include <stdio.h>
#include <assert.h>
#include "cc2y4m.h"
#ifdef DBG
#include <SDL/SDL.h>
#endif

void lock_init(lock_t *lock, bttv_t *bttv) {
	lock->stabilize = 0;
	lock->fieldcount = 0;
	lock->lastref = -1;
	lock->state = 0;
	lock->fframe = 0;
	lock->lframe = 0;
	lock->errors = 0;
	lock->bttv = bttv;

	pthread_mutex_init(&lock->avail, NULL);
	pthread_mutex_init(&lock->avail2, NULL);
	pthread_cond_init(&lock->cavail, NULL);
}

/* Methods for locking detection */

/* this routine tries to find:
 * - a black (<= TRESHOLD) line with a minimal length of 50 followed by
 * - a white (> TRESHOLD) line with a length between 335 and 365
 *
 * If precisely one of these things is found this routine reports the
 * start point of the while line.
 *
 * There is no garuantee that it returns the correct sync point,
 * but there is a high enough probability... */

int refpoint_m1(lock_t *lock, unsigned char *buf, int start, int stop) {
	int hits, state, i, count, hitpoint, prevlen = 0;
	state = ( *buf > method->locktresh ) ? 1 : 0;
	count = 0;
	hitpoint = -1;
	hits = 0;
	if (start < 0) start = 0;
	/* quick search */
	if (stop - start < 40) {
		for (i = start; i < stop; i++) if (*(buf+i) > method->locktresh) 
			return i-1;
		return -1;
	}

	for (i = start; i < stop; i++) {
		if (state == ( *(buf + i) > method->locktresh ) ? 1 : 0) count++;
		else {
			if (count > 335 && count < 365 && state == 1 && 
					prevlen > 50) {
				/* potential hit */
				hits++;
				hitpoint = i - count;
			}
			prevlen = count;
			count = 0;
			state = 1 - state;
		}
	}
	if (hits == 1) return hitpoint;
	else return -1;
}


/* This method tries to find 12 or more consecutive white lines of width>=1880 
 *
 * If the first line is at position l,c the others will be at positions
 * l+1,c l+2,c etc.
 *
 * So if the first line starts at position i the others will start at
 * postion i+3*756, i+2(3*756), i+ 3(3*756) etc.
 */
int refpoint_m2(lock_t *lock, unsigned char *buf, int start, int stop) {
	int i, count = 0;

	if (start < 0) start = 0;

	/* quick search */
	if (stop - start < 40) {
 		for (i = start; i < stop; i++) if (*(buf+i) > method->locktresh) 
				return i-1;
		return -1;
	}

	for (i = start; i < stop; i++) {
		if (*(buf + i) > method->locktresh && count<1880) count++;
		else { 
			if (count >= 1880) return (i-count)+10*3*756;
			count = 0;
		}
  	}
	return -1;
}



double runavg(double addfrq, double cstfrq, int frno, 
		int *fframe, int *lframe) {
	double new = (cstfrq*(*lframe - *fframe) + 
			addfrq*(frno - *lframe))/(frno - *fframe);
	*lframe = frno;
	if (frno - *fframe > 150) *fframe = frno - 150;
	return new;
}

/* INVIS is the number of invisible pixels after a PLL change
 * just after v4l sync; an experimentally derived constant,
 * can be improved, but it probably won't matter
 * anyway, 1417500 is the number of visible+invisible
 * pixels in a frame (625*2268) */
#define INVIS 70000
static int flutlock(lock_t *lock, char *frame, int frno) {
	int ref, predict;
	double speed, factor;
	switch (lock->state) {
		case 0:
			if ((ref = method->refpoint(lock, frame, 0, 2268*288)) 
					>= 0) {
				lock->lastref = ref;
				lock->fframe = lock->lframe = frno;
				lock->state = 1;
			} else lock->errors++;
			ref = -1;
			break;
		case 1:
			if ((ref = method->refpoint(lock, frame, 0, 2268*288)) 
					>= 0) {
				speed = lock->bttv->finfos[frno%10].pll/(2268*625+(double)(ref-lock->lastref)/((double)(frno - lock->lframe)));
				if (speed < 25.006 && speed > 24.994) {
					lock->avgrate = speed;
					lock->fframe = frno;
					lock->lframe = frno;
					lock->lastref = ref;
					lock->state = 2;
					lock->errors = 0;
				} else lock->state = 0;
			} else lock->errors++;
			ref = -1;
			break;
		case 2:
			factor = (1417500-(double)lock->lastref-INVIS)/1417500;
			predict = lock->lastref + ((factor*lock->bttv->finfos[(frno-1)%10].pll + (1-factor)*lock->bttv->finfos[frno%10].pll)/lock->avgrate - 2268*lock->bttv->finfos[(frno-1)%10].vtotal);
			ref = method->refpoint(lock, frame, predict - 10, predict + 10);
			//printf("%d\n", frno - lock->lframe);
			if (ref != -1) {
				//printf("err=%d\n", ref - predict);
				speed = (factor*lock->bttv->finfos[(frno-1)%10].pll + (1-factor)*lock->bttv->finfos[frno%10].pll)/(2268*lock->bttv->finfos[(frno-1)%10].vtotal+(double)(ref-lock->lastref));
				if (speed < 25.006 && speed > 24.994) {
					lock->avgrate = runavg(speed, lock->avgrate, frno, &lock->fframe, &lock->lframe);
					lock->lastref = ref;
					lock->errors = 0;
				} else {
					lock->errors++;
					lock->lastref = predict;
					lock->lframe = frno;
					ref = predict;
				}
			} else {
				lock->errors++;
				lock->lastref = predict;
				lock->lframe = frno;
				ref = predict;
			}
			break;
		default:
			assert(0);
	}
	if (lock->errors == 6) {
		ERROR("trouble syncing... retrying...\n");
		lock->errors = 0;
		lock->state = 0;
		if(method->refpoint == refpoint_m2)
			bttv_schedule_updown(lock->bttv, 2);	
		else if (method->refpoint == refpoint_m1) 
			bttv_schedule_updown(lock->bttv, -2);	
		return -1;
	}
	return ref;
}

void lock_thread(void *data) {
	lock_t *lock = data;
	double frq;
	int frno;
	while (1) {
		pthread_testcancel();
		pthread_mutex_lock(&lock->avail);
		lock->in = bttv_getnext(lock->bttv, &frno);
		lock->offset = flutlock(lock, lock->in, frno);
		//DEBUG("offset=%d\n", lock->offset);
		if (lock->offset >= 0) {
			frq = 2268*625*lock->avgrate;
			bttv_schedule_pll(lock->bttv, frq);
			if (lock->offset/2268 > 6) {
				lock->stabilize = 0;
				bttv_schedule_updown(lock->bttv, 2);
			} else if (lock->offset/2268 > 3) {
				lock->stabilize = 0;
				bttv_schedule_updown(lock->bttv, 1);
			} else if (lock->offset/2268 > 0 && ++lock->stabilize%4 == 0)
				bttv_schedule_updown(lock->bttv, 1);
			if (lock->offset%2268 > method->center+300) {
				bttv_schedule_pll(lock->bttv, frq-1000);
			} else if (lock->offset%2268 > method->center+100) {
				bttv_schedule_pll(lock->bttv, frq-300);
			} else if (lock->offset%2268 > method->center+10) {
				bttv_schedule_pll(lock->bttv, frq-72);
			} else if (lock->offset%2268 > method->center+5) {
				bttv_schedule_pll(lock->bttv, frq-36);
			} else if (lock->offset%2268 > method->center+1) {
				bttv_schedule_pll(lock->bttv, frq);
			} else if (lock->offset%2268 < method->center-300) {
				bttv_schedule_pll(lock->bttv, frq+1000);
			} else if (lock->offset%2268 < method->center-100) {
				bttv_schedule_pll(lock->bttv, frq+300);
			} else if (lock->offset%2268 < method->center-10) {
				bttv_schedule_pll(lock->bttv, frq+72);
			} else if (lock->offset%2268 < method->center-5) {
				bttv_schedule_pll(lock->bttv, frq+36);
			} else if (lock->offset%2268 < method->center-1) {
				bttv_schedule_pll(lock->bttv, frq);
			}
		}
		lock->offset -= method->center;
		pthread_mutex_unlock(&lock->avail);
		pthread_mutex_lock(&lock->avail2);
		pthread_cond_signal(&lock->cavail);
		//pthread_cond_wait(&lock->cavail, &lock->avail2);
		pthread_mutex_unlock(&lock->avail2);
	}
}

char *lock_getnext(lock_t *lock, int *offset) {
	char *buf;
	pthread_mutex_lock(&lock->avail2);
	pthread_cond_wait(&lock->cavail, &lock->avail2);
	buf = lock->in;
	*offset = lock->offset;
	pthread_mutex_unlock(&lock->avail2);
	return buf;
}

void lock_finished(lock_t *lock) {
	pthread_mutex_lock(&lock->avail2);
	pthread_cond_signal(&lock->cavail);
	pthread_mutex_unlock(&lock->avail2);
}
